using System;
using System.Collections;
using System.Reflection;
namespace Stubgen
{
	public class TreeGen
	{
		BindingFlags usefulMembers=BindingFlags.Public | BindingFlags.Static 
				| BindingFlags.Instance | BindingFlags.NonPublic;
		Options options;
		
		Assembly currentAssembly=null;

		private void WriteLine(String fmt,params Object []args)
		{
			String str=String.Format(fmt,args);
			options.Out.WriteLine(str.Replace("\t","  "));
		}
		
		public TreeGen(Options options)
		{
			this.options=options;
		}

		public void WritePreamble()
		{
			WriteLine("<Libraries>");
			WriteLine("<Types Library=\"something\">");
		}

		public void WritePostamble()
		{
			WriteLine("</Types>");
			WriteLine("</Libraries>");
		}

		private String TypeAttributesToCsharp(TypeAttributes attr)
		{
			String str="";
			if((attr & TypeAttributes.Public)!=0) str+="public ";
			if((attr & TypeAttributes.Abstract)!=0) str+="abstract ";
			if((attr & TypeAttributes.Sealed)!=0) str +="sealed ";
			return str.Trim();
		}

		private String TypeParentsToCsharp(Type type)
		{
			String str=": ";
			if(type.BaseType!=null && type.BaseType!=typeof(System.Object))
			{
				str+=type.BaseType.Name+",";
			}
			foreach(Type each in type.GetInterfaces())
			{
				str+=each.Name+",";
			}			
			if(str==": ")return "";
			return str.Substring(0,str.Length-1);
		}

		public String MethodSignatureToCsharp(MethodInfo method)
		{
			String paramsList="(";
			foreach(ParameterInfo pinfo in method.GetParameters())
			{
				paramsList+=pinfo.ParameterType.Name+" "+pinfo.Name+",";
			}
			if(paramsList!="(")
			{
				paramsList=paramsList.Substring(0,paramsList.Length-1);
			}
			paramsList+=")";
			return paramsList;
		}
	
		public void ProcessMember(MemberInfo member)
		{
		}

		public void ProcessEnumMembers(Type type)
		{
			foreach(MemberInfo minfo in type.GetMembers(usefulMembers))
			{
			}
		}

		public void ProcessType(Type type)
		{	
			if(!type.IsPublic)return;
			WriteLine("<Type Name=\"{0}\" FullName=\"{1}\" FullNameSP=\"{2}\">",type.Name,type.FullName,type.FullName);
			if(type.IsEnum)
			{
				WriteLine("\t<TypeSignature Language=\"C#\" Value=\"{0} enum {1}\"/>",TypeAttributesToCsharp(type.Attributes),type.Name);
				ProcessEnumMembers(type);
			}
			else if(type.IsInterface)
			{
				WriteLine("\t<TypeSignature Language=\"C#\" Value=\"{0} interface {1}{2}\"/>",
				TypeAttributesToCsharp(type.Attributes & 
									~(TypeAttributes.Abstract)),
				type.Name,
				TypeParentsToCsharp(type));
			}
			else if(type.IsValueType)
			{
				WriteLine("\t<TypeSignature Language=\"C#\" Value=\"{0} struct {1}\"/>",
				TypeAttributesToCsharp(type.Attributes),
				type.Name);
			}
			else if(type!=typeof(System.Delegate) && 
					type!=typeof(System.MulticastDelegate) &&
					type.IsSubclassOf(typeof(System.Delegate)))
			{
				MethodInfo invokeMethod=type.GetMethod("Invoke");
				WriteLine("\t<TypeSignature Language=\"C#\" Value=\"{0} delegate {1} {2} {3}\"/>",
				TypeAttributesToCsharp(type.Attributes),
				invokeMethod.ReturnType.Name,
				type.Name,
				MethodSignatureToCsharp(invokeMethod)
				);
			}
			else
			{
				WriteLine("<TypeSignature Language=\"C#\" Value=\"{0} class {1}{2}\"/>",
				TypeAttributesToCsharp(type.Attributes),
				type.Name,
				TypeParentsToCsharp(type));
			}
			WriteLine("\t<Base>");
			if(type.BaseType!=null)
			{
				WriteLine("\t\t<BaseTypeName>{0}</BaseTypeName>",
				type.BaseType.FullName);
			}
			WriteLine("\t</Base>");
			if(type.GetInterfaces()!=null)
			{
				WriteLine("\t<Interfaces>");
				foreach(Type iface in type.GetInterfaces())
				{
					WriteLine("\t\t<Interface>");
					WriteLine("\t\t\t<InterfaceName>{0}</InterfaceName>",iface.FullName);
					WriteLine("\t\t</Interface>");
				}
				WriteLine("\t</Interfaces>");
			}
			WriteLine("</Type>");
		}
	
		public void Generate(String fname)
		{
			Assembly assem=Assembly.Load(fname);
			currentAssembly=assem;
			if(assem == null)
			{
				throw new Exception("Could not load "+fname);
			}
			foreach(Type type in assem.GetTypes())
			{
				ProcessType(type);
			}
		}
	}
}
